#pragma once

#ifndef _CONCURRENT_TRIE_HPP
#define _CONCURRENT_TRIE_HPP

#include <atomic>
#include <cstdint>
#include <vector>
#include <queue>
#include <mutex>
using namespace std;

template <class T>
static constexpr int lsb(T i) {
	static_assert(is_same<T, uint32_t>::value || is_same<T, uint64_t>::value);
	if constexpr(is_same<T, uint32_t>::value)
		return __builtin_ffs(i) - 1;
	else
		return __builtin_ffsll(i) - 1;
}

template <class T>
static constexpr int clz(T i) {
	static_assert(is_same<T, uint32_t>::value || is_same<T, uint64_t>::value);
	if constexpr(is_same<T, uint32_t>::value)
		return __builtin_clz(i);
	else
		return __builtin_clzll(i);
}

template <class T = uint32_t, bool chunked = true>
class concurrent_trie {
public:
	static constexpr T chunk_size = sizeof(T) * CHAR_BIT;
	static constexpr T chunk_mask = chunk_size - 1;
	static constexpr T chunk_bits = lsb(chunk_size);
	static constexpr T allset = ~(T)0;
	static constexpr T one = 1;
	struct chunk {
		T bits;
		T version;
		chunk same_copy() const {
			return {bits, version + 1};
		}
		chunk cleared_copy(int offset) const {
			return {bits & ~(one << offset), version + 1};
		}
		chunk set_copy(int offset) const {
			return {bits | (one << offset), version + 1};
		}
	};
	T N = 0;
private:
	int maxDepth = -1;
	T num_internal_nodes = 0;
	// static_assert(atomic<chunk>::is_always_lock_free);
	vector<atomic<chunk>> chunks;
	static_assert(atomic<T>::is_always_lock_free);
	vector<atomic<T>> sizes;
public:
	concurrent_trie() {}
	T getChild(T i, T c) {
		return (i << chunk_bits) + c + 1;
	}
	T getParent(T i) {
		return (i - 1) >> chunk_bits;
	}
	concurrent_trie(T N) : N(N), maxDepth(N > 1 ? (chunk_size - clz(N - 1) - 1) / chunk_bits : -1 + N), num_internal_nodes(maxDepth >= 0 ? ((one << (maxDepth * chunk_bits)) - 1) / (chunk_size - 1) : 0), chunks(num_internal_nodes + (N + chunk_size - 1) / chunk_size), sizes(chunks.size()) {
		auto level_start = num_internal_nodes;
		auto level_granularity = chunk_size;
		auto count_granularity = level_granularity >> chunk_bits;
		for(int d = maxDepth; d >= 0; d--) {
			for(T i = 0; (i + 1) * level_granularity < N; i++) {
				auto s = level_start + i;
				chunks[s].store({allset, 0});
				if constexpr(chunked)
					sizes[s].store(count_granularity);
				else
					sizes[s].store(level_granularity);
			}
			auto i = (N - 1) / level_granularity;
			auto s = level_start + i;
			auto N_ = N - i * level_granularity;
			auto sz = (N_ + count_granularity - 1) / count_granularity;
			chunks[s].store({allset >> (chunk_size - sz), 0});
			if constexpr(chunked)
				sizes[s].store((N_ + chunk_size - 1) >> chunk_bits);
			else
				sizes[s].store(N_);
			count_granularity = level_granularity;
			level_granularity <<= chunk_bits;
			level_start = getParent(level_start);
		}
	}
	void push(T i) {//first | then increment size
		auto offset = i & chunk_mask;
		auto s = num_internal_nodes + (i >> chunk_bits);
		chunk n = chunks[s].load();
		while(!chunks[s].compare_exchange_weak(n, n.set_copy(offset)));
		bool flag;
		if constexpr(chunked)
			flag = n.bits == 0;
		else
			flag = (n.bits & (one << offset)) == 0;
		if(flag) {
			sizes[s]++;
			for(int d = maxDepth - 1; d >= 0; d--) {
				s = getParent(s);
				i >>= chunk_bits;
				offset = i & chunk_mask;
				n = chunks[s].load();
				while(!chunks[s].compare_exchange_weak(n, n.set_copy(offset)));
				sizes[s]++;
			}
		}
	}
	auto pop(T &sz) { // first decrement size at root, first decrement lower level size then & on current level
		T s = 0;
		if(maxDepth >= 0) {
			sz = sizes[0].load();
			do {
				if(sz == 0)
					break;
			} while(!sizes[0].compare_exchange_weak(sz, sz - 1));
			if(sz > 0) { // guaranteed to take one chunk by induction hypothesis
				T child;
				for(int d = 0; d < maxDepth; d++) {
					int offset;
					T child_sz;
					auto n = chunks[s].load();
					for(;;) {
						offset = lsb(n.bits);
						assert(offset >= 0);
						child = getChild(s, offset);
						child_sz = sizes[child].load();
						do {
							if(child_sz == 0)
								break;
						} while(!sizes[child].compare_exchange_weak(child_sz, child_sz - 1));
						if(child_sz >= 1)	// expect this to hold sooner or later because of inductive hypothesis
							break;
						chunks[s].compare_exchange_weak(n, n.cleared_copy(offset));
					}
					s = child;
				}
			}
		}
		else
			sz = 0;
		return [this, s](chunk &n) -> T { // if t has only 1 bit set, then next time will be empty
			int offset;
			{
				n = chunks[s].load();
				do {
					offset = lsb(n.bits);
					assert(offset >= 0); // after the changes, this should hold!
				} while(!chunks[s].compare_exchange_weak(n, n.cleared_copy(offset)));
			}
			auto i = s - num_internal_nodes;
			if constexpr(!chunked)
				n = {1, 0};
			return (i << chunk_bits) + offset;
		};
	}
};

template <class T = uint32_t>
class concurrent_pq {
public:
	static constexpr T chunk_size = sizeof(T) * CHAR_BIT;
	static constexpr T chunk_mask = chunk_size - 1;
	static constexpr T chunk_bits = lsb(chunk_size);
	static constexpr T allset = ~(T)0;
	static constexpr T one = 1;
	struct chunk {
		T bits;
		chunk cleared_copy(int offset) const {
			return {bits & ~(one << offset)};
		}
		chunk set_copy(int offset) const {
			return {bits | (one << offset)};
		}
	};
	T N = 0;
private:
	T num_chunks = 0;
	priority_queue<T, vector<T>, greater<T>> chunk_q;
	vector<atomic<chunk>> chunks;
	mutex mtx;
public:
	concurrent_pq() {}
	concurrent_pq<T> & operator = (concurrent_pq<T> &&pq) {
		swap(num_chunks, pq.num_chunks);
		swap(chunk_q, pq.chunk_q);
		swap(chunks, pq.chunks);
		return *this;
	}
	concurrent_pq(T N) : N(N), num_chunks((N + chunk_size - 1) / chunk_size), chunk_q(greater<T>(), [&]() {
		vector<T> v(N);
		iota(v.begin(), v.end(), 0);
		return v;
	}()), chunks(num_chunks) {
		for(auto &n: chunks)
			n.store({allset});
		if(N & chunk_mask)
			chunks.back().store({(one << (N & chunk_mask)) - 1});
	}
	void push(T i) {
		auto s = i >> chunk_bits;
		auto offset = i & chunk_mask;
		chunk n = chunks[s].load();
		while(!chunks[s].compare_exchange_weak(n, n.set_copy(offset)));
		if((n.bits & (one << offset)) == 0) {
			lock_guard lock(mtx);
			chunk_q.push(i);
		}
	}
	auto pop(T &sz) {
		T s = 0;
		{
			lock_guard lock(mtx);
			if((sz = chunk_q.size())) {
				s = chunk_q.top();
				chunk_q.pop();
			}
		}
		return [this, s](chunk &n) -> T { // if t has only 1 bit set, then next time will be empty
			int offset = s & chunk_mask;
			n = chunks[s >> chunk_bits].load();
			while(!chunks[s >> chunk_bits].compare_exchange_weak(n, n.cleared_copy(offset)));
			n.bits = 1;
			return s;
		};
	}
};

template <class T = uint32_t>
class concurrent_chunked_pq {
public:
	static constexpr T chunk_size = sizeof(T) * CHAR_BIT;
	static constexpr T chunk_mask = chunk_size - 1;
	static constexpr T chunk_bits = lsb(chunk_size);
	static constexpr T allset = ~(T)0;
	static constexpr T one = 1;
	struct chunk {
		T bits;
		chunk cleared_copy(int offset) const {
			return {bits & ~(one << offset)};
		}
		chunk set_copy(int offset) const {
			return {bits | (one << offset)};
		}
	};
private:
	T num_chunks = 0;
	priority_queue<T, vector<T>, greater<T>> chunk_q;
	vector<atomic<chunk>> chunks;
	mutex mtx;
public:
	concurrent_chunked_pq() {}
	concurrent_chunked_pq<T> & operator = (concurrent_chunked_pq<T> &&pq) {
		swap(num_chunks, pq.num_chunks);
		swap(chunk_q, pq.chunk_q);
		swap(chunks, pq.chunks);
		return *this;
	}
	concurrent_chunked_pq(T N) : num_chunks((N + chunk_size - 1) / chunk_size), chunk_q(greater<T>(), [&]() {
		vector<T> v(num_chunks);
		iota(v.begin(), v.end(), 0);
		return v;
	}()), chunks(num_chunks) {
		for(auto &n: chunks)
			n.store({allset});
		if(N & chunk_mask)
			chunks.back().store({(one << (N & chunk_mask)) - 1});
	}
	void push(T i) {
		auto s = i >> chunk_bits;
		auto offset = i & chunk_mask;
		chunk n = chunks[s].load();
		while(!chunks[s].compare_exchange_weak(n, n.set_copy(offset)));
		if(n.bits == 0) {
			lock_guard lock(mtx);
			chunk_q.push(s);
		}
	}
	auto pop(T &sz) {
		T s = 0;
		{
			lock_guard lock(mtx);
			if((sz = chunk_q.size())) {
				s = chunk_q.top();
				chunk_q.pop();
			}
		}
		return [this, s](chunk &n) -> T { // if t has only 1 bit set, then next time will be empty
			int offset;
			{
				n = chunks[s].load();
				do {
					offset = lsb(n.bits);
					assert(offset >= 0); // after the changes, this should hold!
				} while(!chunks[s].compare_exchange_weak(n, n.cleared_copy(offset)));
			}
			auto i = s;
			return (i << chunk_bits) + offset;
		};
	}
};

#endif