#pragma once

#ifndef _GRAPH_HPP
#define _GRAPH_HPP

#include <memory>
#include <vector>
#include <numeric>
#include <algorithm>
#include <string>
#include <fstream>
#include <cassert>
#include <iterator>
#include <numeric>

#include "phmap.h"
using phmap::flat_hash_set;

#include "utils.hpp"

template<class vIDType, class eCountType = std::size_t>
class graph {
public:
	std::unique_ptr<eCountType[]> xadj;
	std::unique_ptr<vIDType[]> adj;
	vIDType N = 0;
	graph() {}
	graph(std::string filename) {
		std::ifstream raw(filename, std::ios::binary);
		raw.read(reinterpret_cast<char *>(&N), sizeof(N));
		raw.seekg(sizeof(eCountType), std::ios::cur);
		xadj = std::unique_ptr<eCountType[]>(new eCountType[N + 1]);
		raw.read(reinterpret_cast<char *>(xadj.get()), sizeof(eCountType) * (N + 1));
		adj = std::unique_ptr<vIDType[]>(new vIDType[xadj[N]]);
		raw.read(reinterpret_cast<char *>(adj.get()), sizeof(vIDType) * xadj[N]);
	}
	graph(std::vector<vIDType> _outdegs) {
		std::vector<eCountType> outdegs(_outdegs.begin(), _outdegs.end());
		N = outdegs.size();
		xadj = std::unique_ptr<eCountType[]>(new eCountType[N + 1]);
		xadj[0] = 0;
		partial_sum(outdegs.begin(), outdegs.end(), xadj.get() + 1);
		adj = std::unique_ptr<vIDType[]>(new vIDType[xadj[N]]);
	}
	static std::vector<vIDType> convertToOutdegs(std::vector<std::vector<vIDType>> G) {
		std::vector<vIDType> outdegs(G.size());
		std::transform(G.begin(), G.end(), outdegs.begin(), [](const auto &v) {
			return v.size();
		});
		return outdegs;
	}
	graph(std::vector<std::vector<vIDType>> G) : graph(convertToOutdegs(G)) {
		eCountType i = 0;
		for(const auto &v: G)
			for(const auto j: v)
				adj[i++] = j;
	}
	const vIDType size() const {
		return N;
	}
	auto operator [](vIDType i) const {
		return adj.get() + xadj[i];
	}
	void dump(std::string filename) const {
		std::ofstream raw(filename, std::ios::binary);
		raw.write((char *)(&N), sizeof(N));
		raw.write((char *)(xadj.get() + N), sizeof(eCountType));
		raw.write((char *)(xadj.get()), sizeof(eCountType) * (N + 1));
		raw.write((char *)(adj.get()), sizeof(vIDType) * xadj[N]);
	}
	template <class InputIt>
	void fixCSR(InputIt begin, bool setend = true) {
		bool moove = false;
		for(vIDType i = 0; i < N; i++) {
			if(moove) {
				auto oxadj_i = xadj[i];
				xadj[i] = xadj[i - 1] + *(begin + i - 1);
				std::copy(adj.get() + oxadj_i, adj.get() + oxadj_i + *(begin + i), adj.get() + xadj[i]);
			}
			else if((eCountType) *(begin + i) < xadj[i + 1] - xadj[i]) {
				moove = true;
			}
		}
		if(setend)
			xadj[N] = xadj[N - 1] + *(begin + N - 1);
	}

	//template <class ExecutionPolicy = std::execution::seq>
	void addBackEdges() {
		auto _xadj = std::unique_ptr<eCountType[]>(new eCountType[N + 1]);
		std::adjacent_difference(xadj.get(), xadj.get() + N + 1, _xadj.get());
		std::for_each(/*ExecutionPolicy, */adj.get(), adj.get() + xadj[N], [&](const vIDType &j) {
			_xadj[j + 1]++;
		});
		std::partial_sum(_xadj.get(), _xadj.get() + N + 1, _xadj.get());
		auto _adj = std::unique_ptr<vIDType[]>(new vIDType[_xadj[N]]);
		auto indices = std::unique_ptr<vIDType[]>(new vIDType[N]);
		std::iota(indices.get(), indices.get() + N, vIDType());
		std::vector<vIDType> b(N);
		std::for_each(/*ExecutionPolicy, */indices.get(), indices.get() + N, [&](const vIDType &i) {
			std::copy(adj.get() + xadj[i], adj.get() + xadj[i + 1], _adj.get() + _xadj[i]);
			for(auto j = xadj[i]; j < xadj[i + 1]; j++) {
				auto k = adj[j];
				_adj[_xadj[k + 1] - ++b[k]] = i;
			}
		});
		xadj = std::move(_xadj);
		adj = std::move(_adj);
	}

	std::vector<vIDType> shellSeq1() const {
		std::vector<vIDType> deg;
		std::adjacent_difference(xadj.get() + 1, xadj.get() + N + 1, std::back_inserter(deg));
		std::vector<flat_hash_set<vIDType>> buckets(size());
		for(vIDType i = 0; i < size(); i++)
			buckets[deg[i]].insert(i);
		std::vector<vIDType> k(size());
		vIDType crMinDeg = 0;
		vIDType coreNumber = 0;
		for(vIDType z = size(); z > 0; z--) {
			while(crMinDeg < size() && buckets[crMinDeg].empty())
				crMinDeg++;
			auto &bucket = buckets[crMinDeg];
			auto i = *bucket.begin();
			bucket.erase(bucket.begin());
			coreNumber = std::max(coreNumber, deg[i]);
			k[i] = coreNumber;
			for(auto j = xadj[i]; j < xadj[i + 1]; j++) {
				auto k = adj[j];
				if(buckets[deg[k]].erase(k)) {
					crMinDeg = std::min(--deg[k], crMinDeg);
					buckets[deg[k]].insert(k);
				}
			}
		}
		return k;
	}

	std::vector<vIDType> shellSeq() const {
		std::vector<vIDType> deg(size());
		for(vIDType i = 0; i < size(); i++)
			deg[i] = xadj[i + 1] - xadj[i];
		auto comp = [&](const int a, const int b) {
			return deg[a] != deg[b] ? deg[a] < deg[b] : a < b;
		};
		std::set<int, decltype(comp)> S(comp);
		for(size_t i = 0; i < size(); i++)
			S.insert(i);
		std::vector<vIDType> k(size());
		vIDType coreNumber = 0;
		while(!S.empty()) {
			auto i = *S.begin();
			S.erase(S.begin());
			coreNumber = std::max(coreNumber, deg[i]);
			k[i] = coreNumber;
			for(auto j = xadj[i]; j < xadj[i + 1]; j++) {
				auto k = adj[j];
				auto it = S.find(k);
				if(it != S.end()) {
					S.erase(it);
					deg[k]--;
					S.insert(k);
				}
			}
		}
		return k;
	}

	friend inline std::ostream& operator <<(std::ostream &os, const graph<vIDType, eCountType> &G) {
		os << "[\n";
		for(vIDType i = 0; i < G.N; i++)
			os << i << ": " << std::vector<vIDType>(G[i], G[i + 1]) << '\n';
		return os << "]";
	}
};


#endif //GRAPH_HPP
