
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   prefix_sum_tools.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Mon Jun 14 13:03:59 2010
 * 
 * @brief  Basic prefix sum operations on arrays
 * 
 * 
 */


#ifndef __PREFIX_SUM_TOOLS_H__
#define __PREFIX_SUM_TOOLS_H__

#include "prefix2d.hpp"
#include "rect_list.hpp"
namespace util
{
/** 
 * Extracts first dimension cut points and puts them in an array. 2nd
 * dimension cuts are also decomposed and put in a vector because 2nd
 * dimension cuts can be variable size in m-way jagged
 * partitions. This method assumes the input rectangle list is
 * generated by a -HORF variant of jagged partitioning algorithm or a
 * rectilinear algorithm.
 *
 *
 * @param[in] currentP Partition to be decomposed
 *
 * @param[out] cuts Cut points of the 1st dim. the lenght of this array
 * must be procX+1 which means top and bottom of the matrix also
 * included (i.e 1,cut1,cut2,...,matrix_1st_dimension_size). This
 * format follows 1D partitioning format with additional 2 cuts. To
 * retrieve 1D partitioning format simply drop cuts[0] and cuts[procX]
 *
 * @param[out] procs procs[i] is the 2nd dimension cut points in the ith row.
 * 
 * @return number of stripes in the currentP.
 */
template <typename T, typename Pr>
int extractXDim(const util::RectList<T,Pr>& currentP, int *cuts, std::vector<int> procs[]);
/** 
 * Extracts first dimension cut points and puts them in an array. 2nd
 * dimension cuts are also decomposed and put in a vector because 2nd
 * dimension cuts can be variable size in m-way jagged
 * partitions. This method assumes the input rectangle list is
 * generated by a -HORF variant of jagged partitioning algorithm or a
 * rectilinear algorithm.
 *
 *
 * @param[in] currentP Partition to be decomposed
 *
 * @param[out] cuts Cut points of the 1st dim. the lenght of this array
 * must be procX+1 which means top and bottom of the matrix also
 * included (i.e 1,cut1,cut2,...,matrix_1st_dimension_size). This
 * format follows 1D partitioning format with additional 2 cuts. To
 * retrieve 1D partitioning format simply drop cuts[0] and cuts[procX]
 * 
 * @return number of stripes in the currentP.
 */
template <typename T, typename Pr>
int extractXDim(const util::RectList<T,Pr>& currentP, int *cuts);
/** 
 * 
 * @brief Assembles 1d prefix sum array
 *
 * @param[in] array Original 1d array instance 
 * @param[out] result Output prefix sum array with result[0]=0 and result[i]=result[i-1]+array[i]
 * @param[in] length Original array's length 
 */
template<typename T>
void prefixsum(const T *array, T *result, int length);



/**
 * @brief 1D representation of a section of a prefix sum array by
 * summing its element row-wise or column-wise.
 *
 * This is a 1d prefix sum array with both the bracket[] notation
 * defined by util::PrefixSumArray1DOfTBracket and interval
 * notation defined by util::PrefixSumArray1DOfTBracket.
 *
 *
 * @param T type of the array
 * @param Pr type of the real array
 * @param row select the orientation of the summation. (e.g. If row is
 * true and the rectangle is of size 10x20, then the equivalent
 * 1darray is of size 10)
 */

template <typename T, typename Pr, bool row>
class Aggreg2Dto1D
{
  ///the 2d prefix sum array which is represented
  const Pr& pr;
  const int xlow;
  const int xhigh;
  const int ylow;
  const int yhigh;
  T precomp;

private:
  ///deactivated since it does not make sense
  Aggreg2Dto1D<T,Pr,row>& operator=(const Aggreg2Dto1D<T,Pr,row>& rhs);
  ///deactivated since it does not make sense
  Aggreg2Dto1D();

public:

  /** 
   * @brief Reduces the 2D matrix into 1D array.
   *
   * Reduces the 2D matrix into 1D array \f$[xl:xh]\times[yl:yh]\f$ in the
   * original array \f$[1:n]\times[1:m]\f$ (where n and m are last elements
   * not sizes).
   *
   * If row is true, there are (xh-xl) elements in the array.
   *
   * @param p prefixSumArray
   * @param xl xlow of submatrix
   * @param xh xhigh of submatrix
   * @param yl ylow of submatrix
   * @param yh yhigh of submatrix
   */
  Aggreg2Dto1D(const Pr& p, int xl, int xh, int yl, int yh);
  /** 
   * Copy constructor for Aggreg2Dto1D
   * 
   */
  Aggreg2Dto1D(const Aggreg2Dto1D<T,Pr,row>& o);
  const T operator[] (int x) const;
  /** 
   * @brief Returns the load in a given interval. 
   *
   * The interval is ]i1,i2] i.e., i1 is excluded, i2 is included.
   * 
   * @param i1 lower interval point
   * @param i2 higher interval point
   * 
   * @return load
   */
  T interval(int i1, int i2) const;
};




template <typename T, bool>
class Prefix2D;
template <typename T, typename Pr, bool row>
class Aggreg2Dto1D;

/**
 * @brief Reduces a 2D matrix with stripes into 1D dimension by
 * returning the load of the most loaded stripe.
 *
 * Reduces a 2D matrix with stripes into 1D dimension. Those stripes
 * are created by partitioning in one dimension. Given any interval on
 * the other dimension, returns the load of the most loaded stripe for
 * that interval.
 *
 * This provides a 1d prefix sum array using the interval notation as
 * defined by util::PrefixSumArray1DOfTInterval.
 * 
 * @param row select the orientation of the summation. (e.g. If row is
 * true and the rectangle is of size 10x20, then the equivalent
 * 1darray is of size 10)
 */
template <typename T, typename Pr, bool row>
class AggregMax2Dto1D
{
  ///number of fixed rows or columns
  int lineCount;
  ///chunks of partitioned lines (can be row-wise or column-wise). For
  ///example: when we parition row-wise, we get line chunks like
  ///chunk[0]={l1,l2,l3} chunk[1]={l4} etc.
  Aggreg2Dto1D<T,Pr,row> **lineSet;
  
  AggregMax2Dto1D();
  const AggregMax2Dto1D& operator=(const AggregMax2Dto1D& a);

public:
  /** 
   * Copy constructor for AggregMax2Dto1D
   * 
   * @param a 
   */
  AggregMax2Dto1D(const AggregMax2Dto1D& a);
  ~AggregMax2Dto1D();
  /** 
   * Constructor
   * 
   * @param psum 2D prefix sum Matrix
   * @param procY number of processors in X or Y dimension (depending on orientation)
   * @param ycuts fixed cut points for X or Y dimension (TODO: format ?)
   */
  AggregMax2Dto1D(const Pr& psum, int procY, int *ycuts);
  /** 
   * @brief Returns the maximum partition in a given interval. (TODO: please document if this is ]i1:i2] or [i1:i2] or ]i1:i2[ or [i1:i2[)
   * 
   * @param left lower interval bound
   * @param right higher interval bound
   * 
   * @return max partition weight in the given interval
   */
  T interval(int left, int right) const;
  /** 
   * AggregMax2Dto1D is built by defining a number of stripes in the
   * first dimension.  This method returns the number of stripes
   * (chunk of rows/columns in the matrix).  
   * @return number of stripes
   */
  int getLineCount();

};

/** 
   * 
   * @brief search for prefixSumBottleneck inside a sorted array between l and h
   * 
   * @param prefixSumArray a monotonic non-decreasing array 
   * @param low lower bound (included)
   * @param high upper bound (included)
   * @param prefixSumBottleneck value to look for (TODO: shouldn't this parameter be a const reference)
   * 
   * @return the largest index $low <= i <= high$ of prefixSumArray where the value prefixSumArray[i] <= prefixSumBottleneck. if prefixSumBottleneck < prefixSumArray[low] the function returns -1
   * 
   */
template <typename T, typename Pr>
int binarySearch(const Pr & prefixSumArray, int low, int high, T prefixSumBottleneck);

/** 
   * 
   * @brief search for the largest interval rooted at base inside
   * prefixSumArray finishing between low and high whose value is less
   * then prefixSumBottlneck using a binary search.
   *
   * uses interval notation
   *
   * @param prefixSumArray a monotonic non-decreasing array per interval
   * @param low lower bound (included)
   * @param high upper bound (included)
   * @param prefixSumBottleneck value to look for
   * @param base beggining of the interval
   * 
   * @return the largest index $low <= i <= high$ of prefixSumArray
   * where the value prefixSumArray.interval(base,i) <=
   * prefixSumBottleneck. if prefixSumBottleneck <
   * prefixSumArray.interval(base,low) the function returns -1
   * 
   */

template <typename T, typename Pr>
int binarySearch_interval(const Pr& prefixSumArray, int low, int high, T prefixSumBottleneck, int base);

  ///TODO: this function is un documented in doxygen
// returns -1 for impossible; or the first i such that prefixSumArray[i]>= prefixSumBottleneck
template <typename T, typename Pr>
int binarySearchLeft(const Pr & prefixSumArray, int low,
			    int high, T prefixSumBottleneck);

}
#include "prefix_sum_tools_impl.hpp"

#endif
