
// This file is part of SPart, Spatially Located Workload Partitioner.

// Copyright (C) 2011, The Ohio State University
// SPart developed by Erdeniz O. Bas, Erik Saule and Umit V. Catalyurek 

// For questions, comments, suggestions, bugs please send e-mail to: 
// Umit V. Catalyurek   catalyurek.1@osu.edu

// SPart is free software; you can redistribute it and/or modify it under
// the terms of the GNU General Public License as published by the Free
// Software Foundation; either version 2 or (at your option) any later
// version.

// SPart is distributed in the hope that it will be useful, but WITHOUT
// ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
// FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License
// for more details.

// You should have received a copy of the GNU General Public License
// along with SPart; if not, write to the Free Software Foundation, Inc.,
// 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

// As a special exception, if other files instantiate templates or use
// macros or inline functions from this file, or you compile this file
// and link it with other works to produce a work based on this file,
// this file does not by itself cause the resulting work to be covered by
// the GNU General Public License. However the source code for this file
// must still be made available in accordance with section (3) of the GNU
// General Public License.

// This exception does not invalidate any other reasons why a work based
// on this file might be covered by the GNU General Public License.

// The full text of the GPL license version 2 is available in "gpl.txt".

/**
 * @file   rect_list.hpp
 * @author erdeniz <erdeniz@bmi.osu.edu>
 * @date   Sun Aug 29 04:17:03 2010
 * 
 * @brief  This file includes rectangle and rectangle list to represent 2D partitions
 * 
 * 
 */

#ifndef __RECT_LIST_H__
#define __RECT_LIST_H__

#include <list>
#include <vector>
#include <iostream>
#include <assert.h>
#include "util/version.hpp"

namespace util
{

  /**
   * @brief a rectangular region (inside a matrix).
   *
   * Partition definition rectangle.  The total partition weight
   * includes elements on boundaries.
   * i.e. \f$\sum_{x\_top\_l<=x<=x\_bot\_r,y\_top\_l<=y<=y\_bot\_r}
   * array[x][y]\f$. NOTE: the first element of the matrix is numbered
   * (1,1) not (0,0)
   */
  struct rectangle
  {
    /// coordinate x - rectangle top left
    int x_top_l;
    /// coordinate y - rectangle top left
    int y_top_l;
    /// coordinate x - rectangle bottom right
    int x_bot_r;
    /// coordinate y - rectangle bottom right
    int y_bot_r;
    
    rectangle();
    
    rectangle(int xl, int xh, int yl, int yh);
    
    /**
     * @returns whether the rectangle is empty or not.
     */
    bool empty() const;
    
    /** 
     * @brief Check if the boundaries of a given rectangle is inside the matrix.
     *
     * @param prefixSum the psa
     * @return whether the partition is in legal bounds or not. 
     */
    template<typename Pr>
    bool valid_bound( const Pr & prefixSum) const;
    /** 
     * @brief computes the load of the rectangle in a prefix sum array.
     *
     * @param prefixSum is the 2D prefix sum array to use in this
     * rectangle's weight calculation
     * 
     * @return The total weight of this rectangle
     * \f$\sum_{x\_top\_l<=x<=x\_bot\_r,y\_top\_l<=y<=y\_bot\_r}
     * array[x][y]\f$. Notice that array refers to the real array not
     * its prefix sum array representation.
     */
    template <typename T, typename Pr>
    T get_load(const Pr& prefixSum) const;

    /**
     * @brief computes the area of the rectangle.
     *
     * @return the number of cells in the rectangle.
     */
    int get_area() const;
  };

  std::ostream& operator<< (std::ostream& out, const rectangle& r);

  /**
   * @brief a list of rectangle inside Pr
   * @param Pr type of Prefix Sum Array it applies to
   * @param T type of data of the prefix sum array
   */

  template <typename T, typename Pr>
  class RectList
  {
  public:
    typedef std::vector<rectangle> container;
    container rectangles;
    const Pr& prefixSum;

    RectList(const Pr &ps);

    RectList(const RectList &rlist);


    /** 
     * @brief adds a rectangle into the list
     *
     * Make partitions with respect to prefix sum
     * if borders are xHigh < xLow or yHigh < yLow it is considered empty not worth to add to rectangle list
     * @param r the rectangle to add to the rectangle list
     */
    void add_rect(const rectangle& r);

    /** 
     * @brief Clones a given RectList into this RectList
     * 
     * @param rlist Source RectList
     */
    void copy_into(const  RectList &rlist);

    /** 
     * @brief Adds a given RectList into this RectList
     * 
     * @param rlist Source RectList
     */
    void add_into(const  RectList &rlist);

    /**
     * @brief empties the list
     *
     */
    void clear();
    
    /**
     * @brief computes the load of the most loaded rectangle.
     * @return the load of the most loaded rectangle
     */
    T get_maximum_load() const;
    /**
     * @brief computes the load of the least loaded rectangle.
     * recall that empty rectangles are never added to the list !
     * @return the load of the least loaded rectangle
     */

    T get_minimum_load() const;  
  
    /** 
     * @brief checks if it is a valid partition of the load matrix.
     *
     * The algorithm is : If all the rectangles are inside borders of rect_list;
     * don't collide each other;
     * and cover the whole matrix; 
     * then the partitioning is valid.
     * @return Whether this RectList is a valid partitioning or not
     */
    bool valid_part() const;
  };

  template <typename T, typename Pr>
  std::ostream& operator<< (std::ostream& out, const util::RectList<T, Pr>& r);
}

#include "rect_list_impl.hpp"
#endif
