#pragma once

#ifndef _SYNCRONIZATION_HPP
#define _SYNCRONIZATION_HPP

#include <atomic>
#include <condition_variable>
#include <mutex>
#include <array>

using namespace std;

class semaphore {
private:
	mutex mtx;
	condition_variable condition;
	int count;
public:
	semaphore(int _count = 0) : count(_count) {}
	void notify() {
		lock_guard lock(mtx);
		count++;
		condition.notify_one();
	}
	void wait() {
		unique_lock lock(mtx);
		condition.wait(lock, [&]() {
			return count > 0;
		});
		count--;
	}
	bool try_wait() {
		lock_guard lock(mtx);
		if(count > 0) {
			count--;
			return true;
		}
		return false;
	}
};

class barrier {
private:
	const int N;
	atomic<int> count;
	array<semaphore, 2> sem;
public:
	barrier(const int _N) : N(_N), count(0), sem{_N, 0} {}
	void arrive_and_wait() {
		sem[0].wait();
		if(++count < N)
			sem[1].wait();
		if(--count > 0)
			sem[1].notify();
		else for(int i = 0; i < N; i++)
			sem[0].notify();
	}
};

class barrier_with_master {
private:
	const int N;
	int count;
	barrier bar;
	semaphore sem;
	mutex mtx;
	const thread::id master_id;
public:
	barrier_with_master(const int _N) : N(_N), count(0), bar{_N + 1}, sem{_N}, master_id(this_thread::get_id()) {}
	void arrive_and_wait() {
		assert(this_thread::get_id() != master_id);
		sem.wait();
		bar.arrive_and_wait();
		lock_guard lock(mtx);
		if(++count == N) {
			sem.notify();
			for(; count > 0; count--)
				sem.notify();
		}
	}
	template <class Callable>
	void arrive_and_wait(Callable f) {
		assert(this_thread::get_id() == master_id);
		{
			lock_guard lock(mtx);
			bar.arrive_and_wait();
			f();
		}
		sem.wait();
	}
};

template <class vIDType, class Callable>
void parallel_for(atomic<vIDType> &si, const vIDType N, const vIDType K, Callable f) {
	for(;;) {
		auto i = si.fetch_add(K);
		if(i >= N)
			break;
		for(auto j = min(i + K, N); i < j; i++)
			f(i);
	}
}

template <class vIDType, class Callable>
void parallel_for_(atomic<vIDType> &si, const vIDType N, const vIDType K, Callable f) {
	for(;;) {
		auto i = si.fetch_add(K);
		if(i >= N)
			break;
		f(i, min(i + K, N));
	}
}

#endif