#pragma once
/**
 * @file    utils.h
 * @author  myozka mfbalin
 * @brief   Implements common utility/helper functions.
 *
 */

#ifndef _UTILS_HPP
#define _UTILS_HPP

#include <vector>
#include <random>
#include <iostream>
#include <fstream>
#include <sstream>
#include <string>
#include <vector>
#include <cstdio>
#include <mutex>
#include <chrono>
#include <set>
#include <cmath>
#include <cstdint>
#include <numeric>
#include <memory>
#include <iterator>

#define DEF_lastProc (nProc - 1)
#define DEF_firstProc 0

typedef long long sizeType;

inline size_t operator "" _z (unsigned long long int i) {
	return i;
}

template <class vIDType>
std::vector<vIDType> histReader(std::string filename) {
	std::ifstream in(filename);
    if (in.fail()) {
        std::cerr << "Cannot open file: " << filename << std::endl;
        exit(EXIT_FAILURE);
    }
	std::vector<vIDType> hist;
	for(vIDType i; in >> i;)
		hist.push_back(i);
	return hist;
}

class timer {
private:
	const std::string name;
	const std::chrono::high_resolution_clock::time_point start;
	std::mutex &mtx;
	std::ostream &os;
public:
	timer(std::string s, std::mutex &_mtx, std::ostream &_os = std::cerr) : name(s), start(std::chrono::high_resolution_clock::now()), mtx(_mtx), os(_os) {
		std::lock_guard lock(mtx);
		os << name << " has started" << std::endl;
	}
	~timer() {
		std::lock_guard lock(mtx);
		os << name << " took " << time() << "s" << std::endl;
	}
	double time() const {
		return std::chrono::duration_cast<std::chrono::duration<double>>(std::chrono::high_resolution_clock::now() - start).count();
	}
};

template <class vIDType, class Generator>
inline vIDType pickSubsetSize(const vIDType minsize, const vIDType maxsize, Generator& gen) {
	#ifdef DEBUG
    if (minsize>maxsize) {
        std::cout << "Error! MinSize>MaxSize in pickSubsetSize." << std::endl;
        return 0;
    }
    #endif
    return std::uniform_int_distribution<vIDType> (std::max(0,minsize),maxsize)(gen);
}
template <class vIDType>
inline std::vector<vIDType> getAllStartIndices(std::vector<vIDType>& all_local_sizes){
    std::vector<vIDType> all_start_indices;
    all_start_indices.reserve(all_local_sizes.size());
    int it = 1; // Processor id iterator
    all_start_indices.push_back(0);
    while (it < (int)all_local_sizes.size()) {
        all_start_indices.push_back(all_local_sizes[it-1]+all_start_indices[it-1]);
        ++it;
    }
    return all_start_indices;
}

template <class T>
inline void printGraph(std::ostream &os, const std::vector<T>& xadj, const std::vector<T>& adj) {
	os << "[\n";
	for(size_t i = 0; i < xadj.size()-1; i++) {
		os << i << ":";
        for (int j=xadj[i]; j < xadj[i+1]; ++j) {
            os << " " << adj[j];
        }
        os << std::endl;
    }
	os << "]" << std::endl;
	return;
}

template <class T>
inline std::ostream& operator <<(std::ostream &os, const std::vector<T> &v) {
    os << "[";
    for(const auto &i: v)
        os << " " << i;
    os << " ]";
    return os;
}

template <class T>
inline std::ostream& operator <<(std::ostream &os, const std::vector<std::vector<T>> &G) {
	os << "[\n";
	for(size_t i = 0; i < G.size(); i++)
		os << i << ": " << G[i] << '\n';
	return os << "]";
}

template <class T>
inline std::string graphsize(std::vector<std::vector<T>> G) {
	size_t edges = 0;
	for(const auto &v: G)
		edges += v.size();
	std::stringstream ss;
	ss << "|V| = " << G.size() << ", |E| = " << edges;
	return ss.str();
}

template<class vIDType, class eCountType>
inline std::string graphsize(std::vector<eCountType> &xadj, std::vector<vIDType> &adj) {
	std::stringstream ss;
	ss << "|V| = " << xadj.size()-1 << ", |E| = " << adj.size();
	return ss.str();
}

template <class vIDType>
inline std::vector<vIDType> histToShellSeq(std::vector<vIDType> &hist) {
	std::vector<vIDType> ss;
	ss.reserve(accumulate(hist.begin(), hist.end(), vIDType()));
	for(vIDType i = 0; i < (vIDType)hist.size(); i++)
        ss.insert(ss.end(), hist[i], i);
    return ss;
}

template <class T>
inline std::vector<T> shellSeqToHist(const std::vector<T> &shellSeq) {
	std::vector<T> hist(1 + *max_element(shellSeq.begin(), shellSeq.end()));
	for(const auto &i: shellSeq)
		hist[i]++;
	return hist;
}

template <class T>
inline std::vector<T> shellSeqToOutdegs(std::vector<T> &k) {
	std::vector<T> outdegs(k.size());
	std::iota(outdegs.begin(), outdegs.end(), T());
	std::transform(outdegs.begin(), outdegs.end(), outdegs.begin(), [&](T i) {
		return std::min(k[i], (T)k.size() - 1 - i);
	});
	return outdegs;
}

template <class vIDType, class RandomIt, class Set, class Generator>
inline void subsetsample(vIDType a, vIDType b, RandomIt begin, RandomIt end, Set &S, Generator &gen, bool needShuffle = true) {
	auto k = std::distance(begin, end);
	auto it = begin;
	for(auto i = b - k; i < b; i++) {
		auto j = std::uniform_int_distribution<vIDType>(a, i)(gen);
		if(S.insert(j).second)
			*(it++) = j;
		else {
			S.insert(i);
			*(it++) = i;
		}
	}
	if(needShuffle)
		std::shuffle(begin, end, gen);
	// testing `erase` efficiency
	//if(flag) // this is worse
	//	S.erase(S.begin(), S.end());
	//else
	while(begin < end)
		S.erase(*(begin++));
}

template <class CountType = std::size_t>
inline CountType C(int n, int k) {
	return (CountType)((std::tgamma(n + 1) / std::tgamma(k + 1) / std::tgamma(n - k + 1)) + 0.5);
}

/*
 * The complexity of the below algorithm is a(a + k)2^(a(a-1)/2)
 */
template <class CountType = std::size_t>
inline CountType numOfShellConfs(int k, int a, int b) {
	CountType cnt = 0;
	std::size_t up_limit = 1;
	up_limit <<= a * (a - 1) / 2;
	std::vector<std::vector<int>> m(a, std::vector<int>(a));
	for(std::size_t i = 0; i < up_limit; i++) {
		auto t = i;
		for(int i = 0; i < a; i++)
			for(int j = i + 1; j < a; j++, t >>= 1)
				m[j][i] = m[i][j] = (int)(t & 1);
		std::vector<int> rows(a, 0), adrows(a, 0);
		for(int i = 0; i < a; i++) {
			rows[i] = std::accumulate(m[i].begin(), m[i].end(), 0);
			adrows[i] = std::accumulate(m[i].begin() + i + 1, m[i].end(), 0);
		}
		std::vector<CountType> rowcnts(a, 0);
		for(int i = 0; i < a; i++)
			for(int j = std::max(0, k - rows[i]); j + adrows[i] <= k && j <= b; j++)
				rowcnts[i] += C(b, j);
		cnt += std::accumulate(rowcnts.begin(), rowcnts.end(), (CountType)1, std::multiplies<CountType>());
	}
	return cnt;
}

template <class CountType = std::size_t>
inline CountType numOfGraphsWithShellHist(std::vector<int> hist) {
	std::vector<int> psum;
	std::partial_sum(hist.begin(), hist.end(), std::back_inserter(psum));
	CountType cnt = 1;
	for(std::size_t i = 0; i < hist.size(); i++)
		cnt *= numOfShellConfs(i, hist[i], psum.back() - psum[i]);
	return cnt;
}

#endif // UTILS_HPP
