/**
 * @file    main.cpp
 * @author  myozka
 * @brief   Implements the graph generation algorithm
 *
 */
// for in/output
#include <iostream>
#include <fstream>
#include <iterator>
#include <vector>
#include <algorithm>
#include <numeric>
#include <cstddef>
// for MPI
#include <mpi.h>
// own includes
#include "io.hpp"
#include "utils.hpp"
#include "mpi_generator.h"

#include "mpi_utils.hpp"
#include "graph.hpp"
#include "jit_debug.h"
#define CERR(msg) std::cerr << msg << std::endl;

#define nextArg(); argv++;argc--;

/**
 * Prints the usage of the program.
 */
void print_usage() {
    CERR("Usage: mpirun -np <#proc> ./pgen [options]");
    CERR("      Optional arguments:");
    CERR("          -i <file>       Path to the input file.");
    CERR("          -o <file>       Output all edges to the given file. (Not written to a file if not given)");
    CERR("          -x <file>       Custom decomposition file. Start node for each processor given in the file.");
    CERR("          -f              Build full CSR");
    CERR("          -v              Verify Cores Decomposition");
    CERR("          -c <int>        Communication method.");
    // CERR("                              -1  - No Communication (used when -d = 999 by default, or can be set when using with custom decomposition file.");
    // CERR("                               0  - Alltoallv");
    // CERR("                               8  - Alltoallv_Split - split to subgroups after first comm round");
    // CERR("                              16  - One-to-one");
    CERR("                              17  - One-to-one - (No comm for count)");
    CERR("                                      ( : default)");
    CERR("          -r <int>        random seed.");
    CERR("                              Default: None");
    CERR("                              Constraints: r > 0");
    // CERR("          -d <int>        Mapping distribution mode, used if no file is given with -x option.");
    // CERR("                              0   - Balanced node distribution.");
    // CERR("                              1   - Balanced edge distribution.  ( : default) (requires alpha)");
    // CERR("                              2   - Relaxed balanced edge distribution. (requires alpha) ");
    // CERR("                              5   - Optimal Load Balance (requires alpha) ");
    // CERR("                            999   - Guaranteed disjoint. (Eliminated Fix Phase)");
    CERR("          -a <int>        alpha: |N|*alpha + |E| is divided as load per process.");
    CERR("                              Default: 3");
    CERR("                              Constraints: a >= 0");
    // CERR("                              Requires -d option to be 1 or 2. Otherwise, ignored.");

    CERR("      Example:");
    CERR("          mpirun -np 4 ./pgen -i hist.txt");
    CERR("                  Generates a graph with shell sequence numbers given by hist.txt and outputs result to the terminal");
}

int main(int argc, char *argv[], char* mvp[])
{
    // DBG_InstallCoreSigHandlers(argv[0], mvp);

    // Init MPI
    MPI_Init(&argc, &argv);

    // get total size of processors and current rank
    int _rank, _nProc;
    MPI_Comm_rank(MPI_COMM_WORLD, &_rank);
    MPI_Comm_size(MPI_COMM_WORLD, &_nProc);
    const int rank = _rank;
    const int nProc = _nProc;

    std::ios::sync_with_stdio(false);

    bool diff_output = false;
    int comm_mode = 17;
    int randomSeed = -1;
    bool verify = false;
    bool buildFullCSR = false;
    int mapping_mode = 1;
    sizeType alpha = 3;
    std::string outfile_name;
    std::string infile_path;
    char* custom_mapping_file = nullptr;
    timeType times = {};
    uint64_t n = 0;

    // forget about first argument (the executable's name)
    nextArg();
    // parse optional parameters
    while(argc > 0 && argv[0][0] == '-') {
        char option = argv[0][1];
        switch (option) {
            case 'o':
                // next parameter is file name
                diff_output = true;
                outfile_name = std::string(argv[1]);
                nextArg();
                break;
            case 'i':
                infile_path = std::string(argv[1]);
                nextArg();
                break;
            case 'd':
                nextArg();
                mapping_mode = atoi(argv[0]);
                break;
            case 'r':
                nextArg();
                randomSeed = atoi(argv[0]);
                break;
            case 'c':
                nextArg();
                comm_mode = atoi(argv[0]);
                break;
            case 'a':
                nextArg();
                alpha = (sizeType)atoi(argv[0]);
                break;
            case 'x':
                // the next argument must be a file path
                nextArg();
                custom_mapping_file = argv[0];
                mapping_mode = -1;
                break;
            case 'f':
                buildFullCSR = true;
                break;
            case 'v':
                verify = true;
                break;
            default:
                print_usage();
                exit(EXIT_FAILURE);
        }
        // iterate to next argument
        nextArg();
    }
    if ( argc != 0 ) {
        print_usage();
        exit(EXIT_FAILURE);
    }
    // check that the mandatory parameters are present
    if ( infile_path.empty() ) {
        print_usage();
        exit(EXIT_FAILURE);
    }

    print_node_distribution();

    std::vector<uint64_t> hist_starti;

    // read elements from file
    int degeneracy = 0;
    if (rank == DEF_firstProc) {
        // histogram array
        hist_starti = histReader<uint64_t>(infile_path);
        degeneracy = hist_starti.size() - 1;
        #ifdef PRINT
            std::cout << "degen: " << degeneracy << std::endl;
        #endif
    }

    // Timing and barrierized
    MPI_Barrier (MPI_COMM_WORLD);

    // start timer
    double t_start, t_end;
    t_start = MPI_Wtime();

    double comm_time_total = 0.0;
    MPI_Bcast(&degeneracy, 1, MPI_INT, DEF_firstProc, MPI_COMM_WORLD);

    hist_starti.resize(degeneracy+2);
    MPI_Bcast(hist_starti.data(), degeneracy+2, MPI_UINT64_T, DEF_firstProc, MPI_COMM_WORLD);

    std::vector<uint64_t> all_local_sizes = getDecomposition(hist_starti, custom_mapping_file, mapping_mode, DEF_firstProc, rank, nProc, alpha);
    comm_time_total = MPI_Wtime() - t_start;

    if (rank == DEF_lastProc){
        std::cout << "all_local_sizes: " << all_local_sizes << std::endl;
    }
    // assert(all_local_sizes[rank] > 0);
    if (all_local_sizes[rank] == 0) {
        std::cout << "Size at rank " << rank << " is ZERO" << std::endl;
        MPI_Abort(MPI_COMM_WORLD, 1);
        return 1;
    }

    // std::exclusive_scan is not implemented in gcc/9.2.0 yet.
    // std::exclusive_scan(hist_starti.begin(), hist_starti.end(), hist_starti.begin(), 0);
    // Alternative for now
    std::vector<uint64_t> hist(hist_starti.begin(), hist_starti.end());
    for (size_t i=1; i < hist_starti.size(); ++i)
        hist_starti[i] += hist_starti[i-1];
    for( size_t i=hist_starti.size()-1; i > 0; --i)
        hist_starti[i] = hist_starti[i-1];
    hist_starti[0] = 0;

    n = hist_starti[hist_starti.size()-1];
    std::vector<uint64_t> all_start_indices = getAllStartIndices(all_local_sizes);

    if (rank == DEF_lastProc){
        std::cout << "all_start_indices: " << all_start_indices << std::endl;
    }
    std::vector<int> local_elements;
    graph <uint64_t, sizeType> retG;

    local_elements = generate_local_vector(all_local_sizes[rank], all_start_indices, hist_starti, rank);

    retG = generateGraphWithShellSequence(local_elements, (uint64_t) local_elements.size(), n, hist_starti, all_start_indices, comm_mode, randomSeed, &times);

    if (retG.size() == 0) {
        std::cout << "RANK " << rank << " Returned empty graph" << std::endl;
        MPI_Abort(MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
    }

    MPI_Barrier(MPI_COMM_WORLD);
    // std::cout << "Generation Completed" << std::endl;

    // get elapsed time in seconds
    t_end = MPI_Wtime();
    times.all = (t_end - t_start);
    times.communications += comm_time_total; // add initial broadcast

    if (diff_output) {
        retG.dump(outfile_name + "___" + std::to_string(rank) + "___.csr");

        std::ofstream f(outfile_name + "-" + std::to_string(rank) + ".direl");
        for (uint64_t i = 0; i < retG.N; ++i) {
            for (sizeType k = retG.xadj[i]; k < retG.xadj[i+1]; ++k) {
                f << (i + all_start_indices[rank]) << ' ' << retG.adj[k] << '\n';
            }
        }
    }

    graph <uint64_t, sizeType> fullG;

    if (buildFullCSR) {
        // TODO: Redistribute nodes to balance the final edge load
        // fullG = redistributeTheGraph(retG, rank, nProc);

        t_start = MPI_Wtime();

        // fullG = fullCSR(fullG, all_start_indices, rank, nProc);

        fullG = fullCSR(retG, all_start_indices, rank, nProc);
        times.fullCSR = MPI_Wtime() - t_start;
        if (verify) {
            distributedVerifier(fullG, hist_starti);
        }
        if (diff_output) {
            std::ofstream f(outfile_name + "-" + std::to_string(rank) + ".undel");
            for (uint64_t i = 0; i < fullG.N; ++i) {
                for (sizeType k = fullG.xadj[i]; k < fullG.xadj[i+1]; ++k) {
                    f << i + all_start_indices[rank] << ' ' << fullG.adj[k] << '\n';
                }
            }
        }
    }

    sizeType overallTotalEdge = 0;
    MPI_Reduce (&retG.xadj[retG.N], &overallTotalEdge, 1, MPI_LONG_LONG, MPI_SUM, DEF_lastProc, MPI_COMM_WORLD);

    // Create type to do gather only once
    MPI_Datatype MPI_TimeType;
    int typecounts[2] = {9,1}; // 9 doubles, 1 int
    MPI_Datatype typetypes[2] = {MPI_DOUBLE, MPI_INT};
    MPI_Aint typeoffsets[2] = {offsetof(timeType, all), offsetof(timeType, countOfCascade)};
    MPI_Type_create_struct(2, typecounts, typeoffsets, typetypes, &MPI_TimeType);
    MPI_Type_commit(&MPI_TimeType);

    timeType* allTimes;
    std::vector<double> all_ph1;
    std::vector<double> all_generator;
    std::vector<double> all_fix;
    std::vector<double> all_subset;
    std::vector<double> all_comm;
    std::vector<double> all_inicomm;
    std::vector<double> all_fixCSR;
    std::vector<double> all_fullCSR;
    std::vector<int> all_coc;

    if (rank == DEF_lastProc) {
        allTimes = new timeType[nProc];
        MPI_Gather(&times, 1, MPI_TimeType, &allTimes[0], 1, MPI_TimeType, DEF_lastProc, MPI_COMM_WORLD);

        all_ph1.resize(nProc); all_generator.resize(nProc); all_fix.resize(nProc);
        all_subset.resize(nProc); all_comm.resize(nProc); all_inicomm.resize(nProc);
        all_fixCSR.resize(nProc); all_fullCSR.resize(nProc); all_coc.resize(nProc);

        for (int i=0; i < nProc; ++i) {
            all_ph1[i] = allTimes[i].phase1;
            all_generator[i] = allTimes[i].generator;
            all_fix[i] = allTimes[i].fix;
            all_subset[i] = allTimes[i].subset;
            all_comm[i] = allTimes[i].communications;
            all_inicomm[i] = allTimes[i].communications + comm_time_total;
            all_fixCSR[i] = allTimes[i].fixCSR;
            all_fullCSR[i] = allTimes[i].fullCSR;
            all_coc[i] = allTimes[i].countOfCascade;
        }
        delete[] allTimes;

        CERR( "rank " << rank << " countOfCascade : " << all_coc );
        CERR( "rank " << rank << " times.generator : " << all_generator );
        CERR( "rank " << rank << " times.communications : " << all_comm );
        CERR( "rank " << rank << " times.inicomm : " << all_inicomm );
        CERR( "rank " << rank << " times.phase1 : " << all_ph1 );
        CERR( "rank " << rank << " times.subset : " << all_subset );
        CERR( "rank " << rank << " times.fix : " << all_fix );
        CERR( "rank " << rank << " times.fixCSR : " << all_fixCSR );
        CERR( "rank " << rank << " times.fullCSR : " << all_fullCSR );

        CERR( "#n " << n << " #e " << overallTotalEdge );
        CERR( "times.all : " << times.all );
        CERR( "times.generator : " << *max_element(all_generator.begin(), all_generator.end()) );
        CERR( "times.communications : " << *max_element(all_comm.begin(), all_comm.end()) );
        CERR( "times.inicomm : " << *max_element(all_inicomm.begin(), all_inicomm.end()) );
        CERR( "times.phase1 : " << *max_element(all_ph1.begin(), all_ph1.end()) );
        CERR( "times.subset : " << *max_element(all_subset.begin(), all_subset.end()) );
        CERR( "times.fix : " <<  *max_element(all_fix.begin(), all_fix.end()) );
        CERR( "times.phase2 : " << times.phase2 );
        CERR( "times.fixCSR : " << *max_element(all_fixCSR.begin(), all_fixCSR.end()) );
        CERR( "times.fullCSR : " << *max_element(all_fullCSR.begin(), all_fullCSR.end()) );

    }
    else {
        MPI_Gather(&times, 1, MPI_TimeType, nullptr, 1, MPI_TimeType, DEF_lastProc, MPI_COMM_WORLD);
    }

    // output time
    if (rank == DEF_lastProc) {
    }
    MPI_Type_free(&MPI_TimeType);
    // finish up
    MPI_Finalize();
    return 0;
}

