#include <iostream>
#include <fstream>
#include <vector>
#include <random>
#include <string>
#include <thread>
#include <cstdint>

#include "utils.hpp"
#include "graph.hpp"

#include "sequential.hpp"
#include "concurrent_trie.hpp"
#include "kcoregenthreads.hpp"

using namespace std;

#define CERR(msg) std::cerr << msg << std::endl;
#define nextArg(); argv++;argc--;

/**
 * Prints the usage of the program.
 */
void print_usage() {

    CERR("Usage: ./kcoregen [options]");
    CERR("      Optional arguments:");
    CERR("          -s\t\tToggle: Run sequential code (Default: false)");
    CERR("          -p <num_threads>\tNumber of threads to use.");
    CERR("				Default: Max available cores");
    CERR("          -a <double>\tAlpha for load distribution");
    CERR("				Default: 4");
    CERR("          -m\t\tToggle: Mutex (Default: false)");
    CERR("          -n\t\tToggle: No numa - NUMA will be disabled (Default: false)");
    CERR("          -l\t\tToggle: Local - Localized optimizations will be used (Default: true)");
    CERR("          -c\t\tToggle: Chunked - Priority queue will be chunked (Default: true)");
    CERR("          -t\t\tToggle: Trie - A trie will be used as the priority queue (Default: true)");
    CERR("          -b\t\tToggle: Sort backwards - Threads will start picking smaller jobs (Default: false)");
    CERR("          -e <unsigned>\tInit constant");
    CERR("				Default: 1024");
    CERR("          -r <unsigned>\tCopy constant");
    CERR("				Default: 32768");

    CERR("          -i <file>\tPath to the input histogram file.");
    CERR("          -o <file>\tOutput CSR (binary) file name. (Not written to a file if not given)");
    CERR("          -v\t\tToggle: Verify Cores Decomposition (Default: false)");
}



int main(int argc, char *argv[]) {

    bool run_sequential = false;
    unsigned num_threads = thread::hardware_concurrency();
    double alpha = 4.0;
    bool nonuma = false;
    bool local = true;
    bool chunked = true;
    bool trie = true;
    bool mutex = false;
    bool sort_backwards = false;
    unsigned init_constant = 1024;
    unsigned copy_constant = 32768;
    std::string parsed_input;
    std::string parsed_output;
    bool verify = false;

    nextArg();
	while(argc > 0 && argv[0][0] == '-') {
	    char option = argv[0][1];
	    switch (option) {
	        case 's':
	            run_sequential = 1 - run_sequential;
	            break;
	        case 'p':
	            nextArg();
	            num_threads = (unsigned) atoi(argv[0]);
	            break;
	        case 'a':
	            nextArg();
	            alpha = atof(argv[0]);
	            break;
	        case 'n':
	            nonuma = 1 - nonuma;
	            break;
	        case 'l':
	            local = 1 - local;
	            break;
	        case 'c':
	            chunked = 1 - chunked;
	            break;
	        case 't':
	            trie = 1 - trie;
	            break;
	        case 'b':
	            sort_backwards = 1 - sort_backwards;
	            break;
	        case 'e':
	            nextArg();
	            init_constant = (unsigned) atoi(argv[0]);
	            break;
	        case 'r':
	            nextArg();
	            copy_constant = (unsigned) atoi(argv[0]);
	            break;
	        case 'm':
	            mutex = 1 - mutex;
	            break;
	        case 'i':
	            nextArg();
	            parsed_input = std::string(argv[0]);
	            break;
	        case 'o':
	            nextArg();
	            parsed_output = std::string(argv[0]);
	            break;
	        case 'v':
	            verify = 1 - verify;
	            break;
	        default:
	            print_usage();
	            exit(EXIT_FAILURE);
	    }
	    // iterate to next argument
	    nextArg();
	}
	if ( argc != 0 ) {
		std::cout << "argc:" << argc << std::endl;
	    print_usage();
	    exit(EXIT_FAILURE);
	}
	// check that the mandatory parameters are present
	if ( parsed_input.empty() ) {
	    print_usage();
	    exit(EXIT_FAILURE);
	}

    CERR("=======================");
	CERR("Processing " << parsed_input);
	CERR("With arguments:");
    CERR("run_sequential: " << run_sequential);
    CERR("num_threads: " << num_threads);
    CERR("alpha: " << alpha);
    CERR("nonuma: " << nonuma);
    CERR("local: " << local);
    CERR("chunked: " << chunked);
    CERR("trie: " << trie);
    CERR("mutex: " << mutex);
    CERR("sort_backwards: " << sort_backwards);
    CERR("init_constant: " << init_constant);
    CERR("copy_constant: " << copy_constant);
    CERR("parsed_input: " << parsed_input);
    CERR("parsed_output: " << parsed_output);
    CERR("verify: " << verify);
    CERR("=======================");

	auto hist = histReader<uint32_t>(parsed_input);

	graph<uint32_t, uint64_t> G;

	if(run_sequential)
		G = sequentialSSGG(hist);
	else if(mutex)
		G = kcoregenthreadscascadingbitset<uint32_t, uint64_t, concurrent_chunked_pq<uint32_t>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, copy_constant)();
	else if(local) {
		if(trie) {
			if(chunked)
				G = kcoregenthreadscascadingbitset<uint32_t, uint64_t, concurrent_trie<uint32_t, true>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, copy_constant)();
			else
				G = kcoregenthreadscascadingbitset<uint32_t, uint64_t, concurrent_trie<uint32_t, false>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, copy_constant)();
		}
		else {
			if(chunked)
				G = kcoregenthreadscascadingbitset<uint32_t, uint64_t, concurrent_chunked_pq<uint32_t>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, copy_constant)();
			else
				G = kcoregenthreadscascadingbitset<uint32_t, uint64_t, concurrent_pq<uint32_t>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, copy_constant)();
		}
	}
	else {
		if(trie) {
			if(chunked)
				G = kcoregenthreadscascadingbitsetold<uint32_t, uint64_t, concurrent_trie<uint32_t, true>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, init_constant);
			else
				G = kcoregenthreadscascadingbitsetold<uint32_t, uint64_t, concurrent_trie<uint32_t, false>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, init_constant);
		}
		else {
			if(chunked)
				G = kcoregenthreadscascadingbitsetold<uint32_t, uint64_t, concurrent_chunked_pq<uint32_t>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, init_constant);
			else
				G = kcoregenthreadscascadingbitsetold<uint32_t, uint64_t, concurrent_pq<uint32_t>>(hist, num_threads, alpha, nonuma, sort_backwards, init_constant, init_constant);
		}
	}

	cout << G.N << ' ' << G.xadj[G.N] << endl;

	if( ! parsed_output.empty() )
		G.dump(parsed_output + "0.csr");

	if( verify ) {
		G.addBackEdges();
		auto Kp = shellSeqToHist(G.shellSeq1());
		auto [it, jt] = mismatch(hist.begin(), hist.end(), Kp.begin(), Kp.end());
		if(it == hist.end() && jt == Kp.end())
			cout << "shell histograms are the same" << endl;
		else
			cout << "ERROR: shell histograms don't match" << endl;
	}

	return 0;
}
