#pragma once

#ifndef _KCOREGENTHREADS_HPP
#define _KCOREGENTHREADS_HPP

/*
 UVC: Apple Computers don't have NUMA
 so define _KCORE_NUMA to make the code more portable
*/
#ifndef __APPLE_CC__
#define _KCORE_NUMA
#endif

#include <sched.h>
#include <errno.h>

#define handle_error_en(en, msg) \
	do { errno = en; perror(msg); exit(EXIT_FAILURE); } while (0)


#ifdef _KCORE_NUMA
#include <numa.h>
#endif

#include <thread>
#include <atomic>
#include <mutex>

#include <cassert>
#include <cstdint>
#include <climits>
#include <vector>
#include <string>
#include <deque>
#include <algorithm>
#include <numeric>
#include <queue>
#include <memory>
#include <utility>
#include <random>

#include "phmap.h"

#include "syncronization.hpp"
#include "concurrent_trie.hpp"
#include "utils.hpp"
#include "graph.hpp"
#include "sequential.hpp"

using namespace std;
using phmap::flat_hash_set;

mutex iomutex;

#ifdef _KCORE_NUMA
int set_thread_affinity(unsigned cpu_id) {
	cpu_set_t cpuset;
	CPU_ZERO(&cpuset);
	CPU_SET(cpu_id, &cpuset);
	int rc = sched_setaffinity(0, sizeof(cpu_set_t), &cpuset);
	if (rc != 0) {
		std::lock_guard lock(iomutex);
		handle_error_en(rc, "sched_setaffinity");
	}
	return rc;
}
#else
int set_thread_affinity(unsigned ) {
	return 0;
}
#endif

template <class vIDType, class eCountType = size_t, class concurrent_pq_type = concurrent_trie<vIDType>>
class kcoregenthreadscascadingbitset {
	const vector<vIDType> hist;
	const unsigned num_threads;
	const double alpha;
	const bool nonuma;
	const bool sort_backwards;
	const vIDType init_constant;
	const eCountType copy_constant;
	const vIDType maxK;
	const timer tim;
	const vector<vIDType> Hist;
	unsigned num_numa_nodes;
	graph<vIDType, eCountType> G;
public:
	kcoregenthreadscascadingbitset(vector<vIDType> hist, const unsigned num_threads = thread::hardware_concurrency(), const double alpha = 4, const bool nonuma = false, const bool sort_backwards = false, const vIDType init_constant = 1 << 10, const eCountType copy_constant = 1 << 15) : hist([&](){
		while(!hist.empty() && hist.back() == 0)
			hist.pop_back();
		assert(!hist.empty() && hist.back() >= hist.size());
		return hist;
	}()), num_threads(num_threads), alpha(alpha), nonuma(nonuma), sort_backwards(sort_backwards), init_constant(init_constant), copy_constant(copy_constant), maxK(hist.size() - 1), tim("all", iomutex), Hist([&]() {
		vector<vIDType> Hist(hist.size() + 1);
		Hist[0] = 0;
		partial_sum(hist.begin(), hist.end(), Hist.begin() + 1);
		return Hist;
	}()), num_numa_nodes([&]() {
		#ifdef _KCORE_NUMA
			return !nonuma && numa_available() != -1 ? numa_num_configured_nodes() : 1;
		#else
			return 1;
		#endif
	}()), G([&]() {
		graph<vIDType, eCountType> G;
		G.N = Hist.back();
		G.xadj = unique_ptr<eCountType[]>(new eCountType[G.N + 1]);
		return G;
	}()) {
		assert(num_threads > 0);
	}
	auto operator () () {
		{
			lock_guard lock(iomutex);
			cerr << "using " << num_threads << " threads" << '\n';
		}

		//vector<atomic<vIDType>> b(N);

		using Generator = mt19937;

		struct Shell {
			const kcoregenthreadscascadingbitset *g;
			vIDType K;
			atomic<unsigned> numa_id;
			graph<vIDType, eCountType> G;
			unique_ptr<vIDType[]> f;
			unique_ptr<vIDType[]> t;
			unique_ptr<atomic<vIDType>[]> ut;
			unique_ptr<atomic<unsigned>[]> subset_flags;
			unique_ptr<atomic<int>[]> vertex_locks;
			unique_ptr<concurrent_pq_type> pq;
			atomic<int> flag{0}, flag1{0}, flag2{0};
			atomic<vIDType> counter{0}, copy_counter{0}, copy_counter1{0};
			atomic<eCountType> copy_counter2{0};
			atomic<eCountType> offset;
			Shell(const kcoregenthreadscascadingbitset &g_g) : g(&g_g), numa_id(g_g.num_numa_nodes) {}
			auto cost() const {
				return g->hist[K] * (g->alpha + K); //vertex + edge cost
			};
			bool operator < (const Shell &other) const {
				return cost() < other.cost();
			}
			auto process(flat_hash_set<vIDType> &Set, Generator &gen, const unsigned thread_id, const unsigned numa_node) {
				auto temp_flag = flag.load();
				while(temp_flag == 0 && !flag.compare_exchange_weak(temp_flag, 1));
				if(temp_flag == 0) {
					G.N = g->hist[K];
					G.xadj = unique_ptr<eCountType[]>(new eCountType[G.N + 1]);
					G.xadj[G.N] = K * (eCountType)G.N - (K == g->maxK ? K * (eCountType)(K + 1) / 2 : (eCountType)0);
					G.adj = unique_ptr<vIDType[]>(new vIDType[G.xadj[G.N]]);
					f = unique_ptr<vIDType[]>(new vIDType[G.N]);
					t = unique_ptr<vIDType[]>(new vIDType[G.N]);
					ut = unique_ptr<atomic<vIDType>[]>(new atomic<vIDType>[G.N]);
					subset_flags = unique_ptr<atomic<unsigned>[]>(new atomic<unsigned>[G.N]);
					vertex_locks = unique_ptr<atomic<int>[]>(new atomic<int>[G.N]);
					pq = make_unique<concurrent_pq_type>(K < g->maxK ? G.N : G.N - K - 1);
					flag++;
				}
				Set.reserve(K);
				while(flag.load() < 2); //Barrier
				flag1--;
				parallel_for(counter, G.N, g->init_constant, [&](vIDType i) {
					G.xadj[i] = K < g->maxK || i < G.N - (K + 1) ? i * (eCountType)K : G.xadj[G.N] - (G.N - 1 - i) * (eCountType)(G.N - i) / 2;
					f[i] = 0;
					t[i] = g->Hist.back();
					ut[i] = 0;
					subset_flags[i] = 0;
					vertex_locks[i] = 0;
				});
				flag1++;
				while(flag1.load() < 0); //Barrier
				size_t num_processed_chunks = 0_z;
				flag2--;
				for(vIDType sz; ; num_processed_chunks++) {
					auto popper = pq->pop(sz);
					if(sz == 0)
						break;
				#ifdef VERBOSE
					if(num_processed_chunks == 0) {
						lock_guard lock(iomutex);
						cerr << "Thread #" << thread_id << " has started working at Shell #" << K << " at: " << g->tim.time() << '\n';
						cerr << "This thread is originally from numa node #" << numa_node << " and this Shell is assigned to numa nodes with #" << numa_id << '\n';
					}
				#endif
					for(typename concurrent_pq_type::chunk n = {3}; n.bits & (n.bits - 1);) {
						const auto i = popper(n);
						// assert(n.bits > 0);
						int temp;
						for(;;) {
							temp = 0;
							if(vertex_locks[i].compare_exchange_weak(temp, -1))
								break;
						}
						if(temp != 0)
							continue;
						
						if(subset_flags[i]++ == 0)
							subsetsample(g->Hist[K] + i + 1, g->G.N, G.adj.get() + G.xadj[i], G.adj.get() + G.xadj[i + 1], Set, gen);
						
						const vIDType ut_i = ut[i];
						if(ut_i == t[i]) {
							vertex_locks[i] = 0;
							continue;
						}
						t[i] = ut_i;
						const auto r = uniform_int_distribution<vIDType>(max(t[i], K) - t[i], K)(gen);
						const vIDType inc = f[i] < r ? 1 : -1;
						const auto [beg, end] = minmax(f[i], r);
						for(auto j = beg; j < end; j++) {
							//b[adj[j + xadj[i]]]++;
							const auto k = G.adj[j + G.xadj[i]] - g->Hist[K];
							if(k < G.N) {
								ut[k] += inc;
								if(k + g->Hist[K] < g->Hist.back() - K - 1)
									pq->push(k);
							}
						}
						f[i] = r;
						vertex_locks[i] = 0;
					}
				}
				flag2++;
				while(flag2.load() < 0); //Barrier
				if(/*numa_node == numa_id.load() &&*/ flag++ == 2) {
					if(K == g->maxK) {
					#ifdef VERBOSE
						timer tim("phase2", iomutex);
					#endif
						for(auto i = G.N - K - 1; i < G.N; i++)
							t[i] = ut[i].load();

						auto tptr = t.get();
						auto fptr = f.get();
						phase2and3(tptr, fptr, G, Set, gen, K, g->Hist[K]);

					#ifdef VERBOSE
						{
							lock_guard lock(iomutex);
							cerr << "times.phase2 : " << tim.time() << '\n';
						}
					#endif
					}
				#ifdef VERBOSE
					timer tim("Thread #"s + to_string(thread_id) + " fixing Shell #" + to_string(K), iomutex);
				#endif
					G.fixCSR(f.get());
					offset = G.xadj[G.N];
					
					// {
					// 	vector<vIDType> u(G.N);
					// 	for(vIDType i = 0; i < G.N; i++) {
					// 		assert(G.xadj[i + 1] - G.xadj[i] <= K);
					// 		assert(G.xadj[i + 1] - G.xadj[i] + u[i] >= K);
					// 		for(auto j = G.xadj[i]; j < G.xadj[i + 1]; j++)
					// 			if(G.adj[j] - g->Hist[K] < G.N)
					// 				u[G.adj[j] - g->Hist[K]]++;
					// 	}
					// }
				}
			#ifdef VERBOSE
				if(num_processed_chunks > 0) {
					lock_guard lock(iomutex);
					cerr << "Thread #" << thread_id << " has quit working at Shell #" << K << " at: " << g->tim.time() << '\n';
				}
			#endif
				return num_processed_chunks;
			}
			auto copyCSR(unsigned thread_id, unsigned numa_node) {
				eCountType work_counter = 0;
				flag1--;
				parallel_for(copy_counter, G.N + 1, g->init_constant, [&](vIDType i) {
					G.xadj[i] += offset;
					work_counter++;
				});
				flag1++;
				while(flag1.load() < 0);
				parallel_for_(copy_counter1, G.N, (vIDType)g->copy_constant, [&](vIDType i, vIDType j) {
					copy(G.xadj.get() + i, G.xadj.get() + j, g->G.xadj.get() + g->Hist[K] + i);
					work_counter++;
				});
				parallel_for_(copy_counter2, G.xadj[G.N] - G.xadj[0], g->copy_constant, [&](eCountType i, eCountType j) {
					copy(G.adj.get() + i, G.adj.get() + j, g->G.adj.get() + G.xadj[0] + i);
					work_counter++;
				});
				return work_counter;
			}
		};

		vector<unsigned> available_thread_ids;

	#ifdef _KCORE_NUMA
		for(int thread_id = 0; available_thread_ids.size() < num_threads && thread_id < numa_num_configured_cpus(); thread_id++) {
			while(thread_id < numa_num_configured_cpus() && !numa_bitmask_isbitset(numa_all_cpus_ptr, thread_id))
				thread_id++;
			if(thread_id >= numa_num_configured_cpus())
				break;
			available_thread_ids.push_back(thread_id);
		}
	#else
		available_thread_ids.resize(num_threads);
		iota(available_thread_ids.begin(), available_thread_ids.end(), 0);
	#endif
		
		vector<unique_ptr<Shell>> shells;
		vIDType k = 0; 
		for(auto n: hist) {
			if(n > 0) {
				shells.emplace_back(make_unique<Shell>(*this));
				shells.back()->K = k;
			}
			k++;
		}
		sort(shells.begin(), shells.end(), [](const unique_ptr<Shell> &a, const unique_ptr<Shell> &b) {
			return *a < *b;
		});
		if(!sort_backwards)
			reverse(shells.begin(), shells.end());

		vector<atomic<size_t>> current_counts(num_numa_nodes + 1);
		for(auto &current_count: current_counts)
			current_count = 0;
		current_counts.back() = shells.size();
		vector<double> costs(num_numa_nodes, 0);
		for(auto &shell: shells) {
			if(current_counts.back().load() + available_thread_ids.size() <= shells.size())
				break;
			auto numa_id = distance(costs.begin(), min_element(costs.begin(), costs.end()));
			costs[numa_id] += shell->cost();
			current_counts[numa_id]++;
			current_counts.back()--;
			shell->numa_id = numa_id;
		}

		vector<atomic<size_t>> current_counters(num_numa_nodes);
		for(auto &current_counter: current_counters)
			current_counter = 0;

		barrier_with_master bar(available_thread_ids.size());
		vector<thread> pool;

		auto thread_func = [&](const unsigned thread_id) {
			set_thread_affinity(thread_id % thread::hardware_concurrency());
			
			unsigned numa_node = 0;
			
		#ifdef _KCORE_NUMA
			if(!nonuma && numa_available() != -1) {
				numa_node = numa_node_of_cpu(thread_id % thread::hardware_concurrency());
				numa_set_preferred(numa_node);
			}
		#endif

		#ifdef VERBOSE
			{
				lock_guard lock(iomutex);
				cerr << "Thread #" << thread_id << " belongs to numa node #" << numa_node << '\n';
			}
		#endif

			random_device rnd;
			Generator gen(rnd());

			flat_hash_set<vIDType> Set;

			for(unsigned numa_node_inc = 0; numa_node_inc < num_numa_nodes; numa_node_inc++) {
				const auto current_numa_node = (numa_node + numa_node_inc) % num_numa_nodes;
				for(flat_hash_set<vIDType> skipped; skipped.size() < current_counts[current_numa_node] + current_counts.back();) {
					auto &shell = *shells[current_counters[current_numa_node]++ % shells.size()];
					auto temp_num_numa_nodes = num_numa_nodes;
					if(!shell.numa_id.compare_exchange_strong(temp_num_numa_nodes, current_numa_node) && temp_num_numa_nodes != current_numa_node)
						continue;
					if(temp_num_numa_nodes == num_numa_nodes) {
						current_counts[current_numa_node]++;
						current_counts.back()--;
					}
					if(shell.process(Set, gen, thread_id, numa_node) <= 0)
						skipped.insert(shell.K);
				}
			}

			bar.arrive_and_wait(); // L1

			for(unsigned numa_node_inc = 0; numa_node_inc < 1 /*num_numa_nodes*/; numa_node_inc++) {
				const auto current_numa_node = (numa_node + numa_node_inc) % num_numa_nodes;
				for(flat_hash_set<vIDType> skipped; skipped.size() < current_counts[current_numa_node] + current_counts.back();) {
					auto &shell = *shells[current_counters[current_numa_node]++ % shells.size()];
					auto temp_num_numa_nodes = num_numa_nodes;
					if(!shell.numa_id.compare_exchange_strong(temp_num_numa_nodes, current_numa_node) && temp_num_numa_nodes != current_numa_node)
						continue;
					if(temp_num_numa_nodes == num_numa_nodes) {
						current_counts[current_numa_node]++;
						current_counts.back()--;
					}
					if(shell.copyCSR(thread_id, numa_node) <= 0)
						skipped.insert(shell.K);
				}
			}
		};

		for(unsigned i = 0; i < num_threads; i++)
			pool.emplace_back(thread_func, i);

		{
			bar.arrive_and_wait([&]() { // L1
				G.xadj[G.N] = 0;
				vector<vIDType> indices(shells.size());
				iota(indices.begin(), indices.end(), 0);
				sort(indices.begin(), indices.end(), [&](vIDType i, vIDType j) {
					return shells[i]->K < shells[j]->K;
				});
				for(auto shell_id: indices) {
					const auto &shell = shells[shell_id];
					shell->offset = G.xadj[G.N];
					G.xadj[G.N] += shell->G.xadj[shell->G.N];
				}
				G.adj = unique_ptr<vIDType[]>(new vIDType[G.xadj[G.N]]);
				
				for(auto &current_counter: current_counters)
					current_counter = 0;
			});

			lock_guard lock(iomutex);
			cerr << "times.generator : " << tim.time() << '\n';
		}

		{
			timer tim("CSR copying phase", iomutex);

			for(auto &thr: pool)
				thr.join();

			lock_guard lock(iomutex);
			cerr << "times.copyCSR : " << tim.time() << '\n';
		}

		{
			lock_guard lock(iomutex);
			cerr << "times.all : " << tim.time() << '\n';
		}
		
		graph<vIDType, eCountType> temp_G;
		swap(temp_G.N, G.N);
		swap(temp_G.xadj, G.xadj);
		swap(temp_G.adj, G.adj);
		return temp_G;
	}
};

template <class vIDType, class eCountType = size_t, class concurrent_pq_type = concurrent_trie<vIDType>>
auto kcoregenthreadscascadingbitsetold(vector<vIDType> hist, const unsigned num_threads = thread::hardware_concurrency(), const double alpha = 4, const bool nonuma = false, const bool sort_backwards = false, const vIDType init_chunk_constant = 64, const vIDType subset_chunk_constant = 64, const bool nofixcsr = false) {
	int init_chunk_constant_log2 = 0;
	while(((vIDType)1 << init_chunk_constant_log2) < init_chunk_constant)
		init_chunk_constant_log2++;

	{
		lock_guard lock(iomutex);
		cerr << "using " << num_threads << " threads" << '\n';
	}
	timer tim("all", iomutex);

	while(!hist.empty() && hist.back() == 0)
		hist.pop_back();
	assert(!hist.empty() && hist.back() >= hist.size());

	vIDType maxK = hist.size() - 1;
	vector<vIDType> Hist(hist.size() + 1);
	Hist[0] = 0;
	partial_sum(hist.begin(), hist.end(), Hist.begin() + 1);

	graph<vIDType, eCountType> G;

	const auto &N = G.N = Hist.back();
	unique_ptr<vIDType[]> shellseq = unique_ptr<vIDType[]>(new vIDType[N]);

	G.xadj = unique_ptr<eCountType[]>(new eCountType[N + 1]);
	vector<eCountType> HistEdges(Hist.size());
	{
		HistEdges[0] = 0;
		for(vIDType k = 0; k < hist.size(); k++)
			HistEdges[k + 1] = HistEdges[k] + hist[k] * k;
		HistEdges.back() -= maxK * (maxK + (eCountType)1) / 2;
		G.adj = unique_ptr<vIDType[]>(new vIDType[HistEdges.back()]);
	}
	G.xadj[N] = HistEdges.back();

	const auto &xadj = G.xadj;
	const auto &adj = G.adj;

	unique_ptr<eCountType[]> f = unique_ptr<eCountType[]>(new eCountType[N + 1]); // one more for prefix sum and fixCSR etc.
	unique_ptr<vIDType[]> t = unique_ptr<vIDType[]>(new vIDType[N]);
	unique_ptr<atomic<vIDType>[]> ut = unique_ptr<atomic<vIDType>[]>(new atomic<vIDType>[N]);
	//vector<atomic<vIDType>> b(N);

	const vIDType num_elems = N - maxK - 1;

	vector<concurrent_pq_type> shell_chunk_qs(hist.size());
	auto cost = [&](vIDType i) {
		return hist[i] * (alpha + i) * (sort_backwards ? -1 : 1); //vertex + edge cost
	};
	vector<vIDType> shells(hist.size());
	iota(shells.begin(), shells.end(), 0);
	sort(shells.begin(), shells.end(), [&](const vIDType &i, const vIDType &j) {
		return cost(i) > cost(j);
	});

	unsigned num_numa_nodes = 1;

#ifdef _KCORE_NUMA
	if (!nonuma && numa_available() != -1)
		num_numa_nodes = numa_num_configured_nodes();
#endif

	vector<atomic<unsigned>> shell_numas(shells.size());
	for(auto &shell_numa: shell_numas)
		shell_numa = num_numa_nodes;

	atomic<vIDType> chunk_counter = 0;
	atomic<vIDType> init_counter = 0;
	atomic<vIDType> subset_counter = 0;
	vector<atomic<size_t>> current_counters(num_numa_nodes);
	for(auto &current_counter: current_counters)
		current_counter = 0;
	atomic<size_t> shift_counter = 0;
	unique_ptr<vIDType[]> fixed_adj;

	barrier_with_master bar(num_threads);
	vector<thread> pool;

	auto thread_func = [&](const unsigned thread_id) {
		set_thread_affinity(thread_id % thread::hardware_concurrency());
			
		unsigned numa_node = 0;
		
	#ifdef _KCORE_NUMA
		if(!nonuma && numa_available() != -1) {
			numa_node = numa_node_of_cpu(thread_id % thread::hardware_concurrency());
			numa_set_preferred(numa_node);
		}
	#endif

		{
			lock_guard lock(iomutex);
			cerr << "Thread #" << thread_id << " belongs to numa node #" << numa_node << '\n';
		}

		random_device rnd;
		mt19937 gen(rnd());
		flat_hash_set<vIDType> Set;
		parallel_for(chunk_counter, (vIDType)shell_chunk_qs.size(), (vIDType)1, [&](vIDType i) {
			shell_chunk_qs[i] = concurrent_pq_type(i < maxK ? hist[i] : hist[i] - i - 1);
		});
		parallel_for_(init_counter, N, num_threads * init_chunk_constant, [&](vIDType i, vIDType j) {
			vIDType k = distance(Hist.begin() + 1, upper_bound(Hist.begin(), Hist.end(), i));
			for(; i < j; i++) {
				while(i >= Hist[k + 1])
					k++;
				shellseq[i] = k;
				f[i] = 0;
				t[i] = N;
				ut[i] = 0;
				if(i < num_elems)
					xadj[i] = HistEdges[k] + (i - Hist[k]) * (eCountType)k;
				else
					xadj[i] = HistEdges[k + 1] - (N - 1 - i) * (eCountType)(N - i) / 2;
			}
		});
		bar.arrive_and_wait(); // L1
		parallel_for_(subset_counter, num_elems, num_threads * subset_chunk_constant, [&](vIDType i, vIDType j) {
			Set.reserve(shellseq[j - 1]);
			for(; i < j; i++)
				subsetsample(i + 1, N, adj.get() + xadj[i], adj.get() + xadj[i + 1], Set, gen);
		});
		bar.arrive_and_wait(); // L2
		for(size_t skip_counter = 0; skip_counter < num_numa_nodes * shells.size();) {
			const auto current_numa_node = (numa_node + skip_counter / shells.size()) % num_numa_nodes;
			auto &current_counter = current_counters[current_numa_node];
			const auto kshell = shells[current_counter++ % shells.size()];
			auto temp_num_numa_nodes = num_numa_nodes;
			if(!shell_numas[kshell].compare_exchange_strong(temp_num_numa_nodes, current_numa_node) && temp_num_numa_nodes != current_numa_node) {
				skip_counter++;
				continue;
			}
			auto &chunk_q = shell_chunk_qs[kshell];
			const auto shell_start = Hist[kshell];
			size_t num_processed_chunks = 0_z;
			for(vIDType sz; ; num_processed_chunks++) {
				auto popper = chunk_q.pop(sz);
				if(sz == 0) {
					if(num_processed_chunks == 0)
						skip_counter++;
					break;
				}
				if(num_processed_chunks == 0) {
					lock_guard lock(iomutex);
					cerr << "Thread #" << thread_id << " has started working at Shell #" << kshell << " at: " << tim.time() << '\n';
					cerr << "This thread is originally from numa node #" << numa_node << " and this Shell is assigned to numa nodes with #" << shell_numas[kshell] << '\n';
				}
				skip_counter -= skip_counter % shells.size();
				for(typename concurrent_pq_type::chunk n = {3}; n.bits & (n.bits - 1);) {
					const auto i = shell_start + popper(n);
					assert(n.bits > 0);
					const vIDType ut_i = ut[i];
					if(ut_i == t[i])
						continue;
					t[i] = ut_i;
					const auto r = uniform_int_distribution<vIDType>(max(t[i], shellseq[i]) - t[i], shellseq[i])(gen);
					const vIDType inc = f[i] < r ? 1 : -1;
					vIDType f_i = f[i];
					const auto [beg, end] = minmax(f_i, r);
					for(auto j = beg; j < end; j++) {
						//b[adj[j + xadj[i]]]++;
						const auto k = adj[j + xadj[i]];
						if(shellseq[i] == shellseq[k]) {
							ut[k] += inc;
							if(k < num_elems)
								chunk_q.push(k - shell_start);
						}
					}
					f[i] = r;
				}
			}
			if(num_processed_chunks > 0) {
				lock_guard lock(iomutex);
				cerr << "Thread #" << thread_id << " has quit working at Shell #" << kshell << " at: " << tim.time() << '\n';
			}
		}
		bar.arrive_and_wait(); // L3
		if(nofixcsr)
			return;
		if(true) {
			int i;
			for(i = 0; (1_z << i) < N; i++) {
				parallel_for_(shift_counter, (size_t)N, 1_z << (i + 1 + max(init_chunk_constant_log2, i) - i), [&](size_t ll, size_t rr) {
					for(size_t l = ll; l < rr; l += 1_z << (i + 1)) {
						auto m = l | (1_z << i);
						//auto r = min(m + (1_z << i), rr);
						if(m < N) {
							// if(f[l] + xadj[l] < xadj[m]) // f holds current block's edge count
							// 	copy(adj.get() + xadj[m], adj.get() + xadj[r], adj.get() + f[l] + xadj[l]);
							f[l] += f[m];
						}
					}
				});
				bar.arrive_and_wait(); // L4
			}
			while(i-- > 0) {
				parallel_for_(shift_counter, (size_t)N, 1_z << (i + 1 + max(init_chunk_constant_log2, i) - i), [&](size_t ll, size_t rr) {
					for(size_t l = ll; l < rr; l += 1_z << (i + 1)) {
						auto m = l | (1_z << i);
						auto r = m + (1_z << i);
						if(r < N)
							f[m] += f[r];
					}
				});
				bar.arrive_and_wait(); // L5
			}
			parallel_for(shift_counter, (size_t)N, (size_t)init_chunk_constant, [&](size_t i) {
				f[N - i] = f[0] - f[N - i];
			});
			bar.arrive_and_wait(); // L6
			parallel_for(subset_counter, N, num_threads * subset_chunk_constant, [&](size_t i) {
				copy(adj.get() + xadj[i], adj.get() + xadj[i + 1], fixed_adj.get() + f[i]);
			});
		}
		else {
			int block_size_log = 0;
			while((1_z << (2 * block_size_log - min(block_size_log, 15))) < N)
				block_size_log++;

			parallel_for_(shift_counter, (size_t)N, 1_z << block_size_log, [&](size_t i, size_t j) {
				graph<vIDType, eCountType> tempG;
				tempG.xadj = unique_ptr<eCountType[]>(xadj.get() + i);
				tempG.adj = unique_ptr<vIDType[]>(adj.get());
				tempG.N = j - i;

				tempG.fixCSR(f.get() + i, false);

				tempG.xadj.release();
				tempG.adj.release();
			});
			bar.arrive_and_wait(); // L4
			parallel_for_(shift_counter, (size_t)N, 1_z << block_size_log, [&](size_t , size_t j) {
				xadj[j] = xadj[j - 1] + f[j - 1];
			});
			bar.arrive_and_wait(); // L5
			unique_ptr<vIDType[]> buffer = unique_ptr<vIDType[]>(new vIDType[((1_z << block_size_log) * maxK + num_threads - 1) / num_threads]);
			for(size_t block_start = 1_z << block_size_log; block_start < N; block_start += 1_z << block_size_log) {
				const auto block_size = min((size_t)N - block_start, 1_z << block_size_log);
				parallel_for_(shift_counter, block_size, (block_size + num_threads - 1) / num_threads, [&](size_t i, size_t j) {
					const auto shift_amount = xadj[block_start + i + 1] - f[block_start + i] - xadj[block_start + i];
					if(shift_amount > 0)
						copy(adj.get() + xadj[block_start + i] + shift_amount, adj.get() + xadj[block_start + j], buffer.get());
					const auto buffer_end = buffer.get() + xadj[block_start + j] - xadj[block_start + i] - shift_amount;
					const auto adj_start = adj.get() + xadj[block_start + i];
					bar.arrive_and_wait(); // L6
					if(shift_amount > 0)
						copy(buffer.get(), buffer_end, adj_start);
					for(; i < j; i++)
						xadj[block_start + i + 1] -= shift_amount;
				});
				bar.arrive_and_wait(); // L7
			}
		}
	};

	for(unsigned i = 0; i < num_threads; i++)
		pool.emplace_back(thread_func, i);

	{
		timer phase1("phase1", iomutex);
		{
			timer tim("initialization phase", iomutex);
			bar.arrive_and_wait([]() {}); // L1
			{
				lock_guard lock(iomutex);
				cerr << "times.init : " << tim.time() << '\n';
			}
		}

		{
			timer tim("subset selection phase", iomutex);
			bar.arrive_and_wait([]() {}); // L2
			{
				lock_guard lock(iomutex);
				cerr << "times.subset : " << tim.time() << '\n';
			}
		}

		{
			timer tim("iterative phase", iomutex);
			bar.arrive_and_wait([&]() { // L3
				{
					f[N] = 0;
					lock_guard lock(iomutex);
					cerr << "times.fix : " << tim.time() << '\n';
					cerr << "times.phase1 : " << phase1.time() << '\n';
				}
				timer tim("phase2", iomutex);
				for(auto i = num_elems; i < N; i++)
					t[i] = ut[i];

				random_device rnd;
				mt19937 gen(rnd());

				flat_hash_set<vIDType> Set;
				Set.reserve(maxK);

				auto tptr = t.get(), fptr = f.get();
				phase2and3(tptr, fptr, G, Set, gen, maxK);

				{
					lock_guard lock(iomutex);
					cerr << "times.phase2 : " << tim.time() << '\n';
				}
			});
		}
	}

	{
		timer tim("CSR fixing phase", iomutex);
		if(!nofixcsr) {
			if(true) {
				timer tim("xadj computation", iomutex);
				int i;
				for(i = 0; (1_z << i) < N; i++)
					bar.arrive_and_wait([&]() { // L4
						shift_counter = 0;
					});
				while(i-- > 0)
					bar.arrive_and_wait([&]() { // L5
						shift_counter = 0;
					});
				bar.arrive_and_wait([&]() { // L6
					f[0] = 0;
					fixed_adj = unique_ptr<vIDType[]>(new vIDType[f[N]]);
					subset_counter = 0;
				});
			}
			else {
				int block_size_log = 0;
				while((1_z << (2 * block_size_log - min(block_size_log, 15))) < N)
					block_size_log++;

				{
					timer tim("L4 & L5", iomutex);
					{
						timer tim("L4", iomutex);
						bar.arrive_and_wait([&]() { // L4
							shift_counter = 0;
						});
					}
					{
						timer tim("L5", iomutex);
						bar.arrive_and_wait([&]() { // L5
							shift_counter = 0;
						});
					}
				}
				{
					timer tim("L6 & L7", iomutex);
					for(size_t block_start = 1_z << block_size_log; block_start < N; block_start += 1_z << block_size_log) {
						{
							//timer tim("L6 -> block_start: "s + to_string(block_start), iomutex);
							bar.arrive_and_wait([&]() {}); // L6
						}
						{
							//timer tim("L7 -> block_start: "s + to_string(block_start), iomutex);
							bar.arrive_and_wait([&]() { // L7
								shift_counter = 0;
							});
						}
					}
				}
			}
			lock_guard lock(iomutex);
			cerr << "parallel fixCSR" << '\n';
		}
		// else {
		// 	G.fixCSR(f.get());
		// 	lock_guard lock(iomutex);
		// 	cerr << "sequential fixCSR" << '\n';
		// }

		for(auto &thr: pool)
			thr.join();

		if(!nofixcsr && true)
			swap(G.adj, fixed_adj);

		{
			lock_guard lock(iomutex);
			cerr << "times.fixCSR : " << tim.time() << '\n';
		}
	}

	{
		lock_guard lock(iomutex);
		cerr << "times.all : " << tim.time() << '\n';
	}
	return G;
}

#endif